/*
 * main.cxx
 *
 * Main source for for PlawWav
 *
 * A H.323 simple answering application
 *
 * Copyright (c) 2005 Nickolay V. Shmyrev <nshmyrev@yandex.ru>
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.0 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific language governing rights and limitations
 * under the License.
 *
 * The Original Code is Portable Windows Library.
 *
 * The Initial Developer of the Original Code is Equivalence Pty. Ltd.
 *
 * Portions are Copyright (C) 1993 Free Software Foundation, Inc.
 * All Rights Reserved.
 *
 * Contributor(s): ______________________________________.
 *
 * $Log: main.cxx,v $
 *
 * Revision 1.3  2006/04/13 05:40:18  nshmyrev
 * Updated to work with openh323-1.18.0 and some options behaviour
 *
 * Revision 1.2  2005/10/01 05:40:18  nshmyrev
 * More clean code
 *
 * Revision 1.1  2005/09/25 00:15:18  nshmyrev
 * Initial version
 *
 */

#include <ptlib.h>
#include <ptlib/pipechan.h>

#include "version.h"
#include "main.h"

PCREATE_PROCESS(PlayWav);

#define new PNEW

#define	DEFAULT_MSG_LIMIT	30
#define DEFAULT_CALL_LOG	"call_log.txt"

static PMutex logMutex;
static PTextFile logFile;
static PFilePath logFilename = DEFAULT_CALL_LOG;

static void LogMessage(const PString & str)
{
  PTime now;
  PString msg = now.AsString("hh:mm:ss dd/MM/yyyy") & str;
  logMutex.Wait();

  if (!logFile.IsOpen()) {
    logFile.Open(logFilename, PFile::ReadWrite);
    logFile.SetPosition(0, PFile::End);
  }

  logFile.WriteLine(msg);

  logFile.Close();
  
  logMutex.Signal();
}

static void LogCall(unsigned len,
		    const PString & addr,
                    const PString & calledParty,
                    const PString & callerParty,
                    const PString & callerID,
                    const PString & transmitCodecName)
{
  LogMessage( PString(PString::Unsigned, len) & 
	      addr & 
	      calledParty &  
	      callerParty & 
	      callerID & 
	      transmitCodecName);
}


///////////////////////////////////////////////////////////////

PlayWav::PlayWav()
  : PProcess("OpenH323 Project", "PlayWav",
             MAJOR_VERSION, MINOR_VERSION, BUILD_TYPE, BUILD_NUMBER)
{
}


PlayWav::~PlayWav()
{
}


void PlayWav::Main()
{
  cout << GetName()
       << " Version " << GetVersion(TRUE)
       << " by " << GetManufacturer()
       << " on " << GetOSClass() << ' ' << GetOSName()
       << " (" << GetOSVersion() << '-' << GetOSHardware() << ")\n\n";

  PConfigArgs args(GetArguments());

  args.Parse(
             "D-disable:"
	     "H-no-hangup."
             "g-gatekeeper:"         "n-no-gatekeeper."
             "h-help."
             "i-interface:"          "-no-interface."
             "-listenport:"         "-no-listenport."
             "m-message:"            "-no-message."
             "l-loop."
#if PTRACING
             "o-output:"
#endif
             "P-prefer:"
             "-port:"
             "r-run:"                "-no-run."
             "-require-gatekeeper."  "-no-require-gatekeeper."
             "-save."
#if PMEMORY_CHECK
             "-setallocationbreakpoint:"
#endif
#if PTRACING
             "t-trace."
#endif
	     "u-username:"           "-no-username."
	     "p-password:"
             ,FALSE);

#if PMEMORY_CHECK
  if (args.HasOption("setallocationbreakpoint"))
    PMemoryHeap::SetAllocationBreakpoint(args.GetOptionString("setallocationbreakpoint").AsInteger());
#endif

#if PTRACING
  PTrace::Initialise(args.GetOptionCount('t'),
                     args.HasOption('o') ? (const char *)args.GetOptionString('o') : NULL);
#endif

  if (args.HasOption('h')) {
    cout << "Usage : " << GetName() << " [options]\n"
            "Options:\n"
            "  -m --message fn     : Set initial outgoing message.\n"
	    "  -r --run            : Set script to generate additional messages.\n"
            "  -H --no-hangup      : don't hangup after playing message\n"
	    "  -l --loop           : Play in a loop\n"
            "  -i --interface ip   : Bind to a specific interface\n"
            "  --listenport port   : Listen on a specific port\n"
            "  -g --gatekeeper host: Specify gatekeeper host.\n"
            "  -u --username str   : Set the local endpoint name to str\n"
	    "  -p --password str   : Set the gatekeeper password to str\n"
            "  -n --no-gatekeeper  : Disable gatekeeper discovery.\n"
            "  --require-gatekeeper: Exit if gatekeeper discovery fails.\n"
            "  -D --disable codec  : Disable the specified codec (may be used multiple times)\n"
            "  -P --prefer codec   : Prefer the specified codec (may be used multiple times)\n"	    
#if PTRACING
            "  -t --trace          : Enable trace, use multiple times for more detail\n"
            "  -o --output         : File for trace output, default is stderr\n"
#endif
            "     --save           : Save arguments in configuration file\n"
            "  -h --help           : Display this help message\n";
    return;
  }

  args.Save("save");

  MyH323EndPoint endpoint;

  PString userName = "Simple Answering Machine v" + GetVersion();
  if (args.HasOption('u')) {
    userName = args.GetOptionString('u');
  }
  endpoint.SetLocalUserName(userName);

  if (args.HasOption('p')) {
    const PString password = args.GetOptionString('p');
    endpoint.SetGatekeeperPassword(password);
  }
  
  if (!endpoint.Initialise(args))
    return;

  // start the H.323 listener
  H323ListenerTCP * listener;
  PIPSocket::Address interfaceAddress(INADDR_ANY);
  WORD listenPort = H323EndPoint::DefaultTcpPort;

  if (args.HasOption("listenport"))
    listenPort = (WORD)args.GetOptionString("listenport").AsInteger();

  if (args.HasOption('i'))
    interfaceAddress = PIPSocket::Address(args.GetOptionString('i'));

  listener  = new H323ListenerTCP(endpoint, interfaceAddress, listenPort);

  if (!endpoint.StartListener(listener)) {
    PTRACE (2, "Could not open H.323 listener port on "
         << listener->GetListenerPort());
    delete listener;
    return;
  }

  if (args.HasOption('g')) {
    PString gkName = args.GetOptionString('g');
    if (endpoint.SetGatekeeper(gkName, new H323TransportUDP(endpoint)))
      PTRACE (2, "Gatekeeper set: " << *endpoint.GetGatekeeper());
    else {
      PTRACE (2, "Error registering with gatekeeper at \"" << gkName << '"');
      return;
    }
  }
  else if (!args.HasOption('n')) {
    PTRACE (2, "Searching for gatekeeper...");
    if (endpoint.DiscoverGatekeeper(new H323TransportUDP(endpoint)))
      PTRACE (2, "Gatekeeper found: " << *endpoint.GetGatekeeper());
    else {
      PTRACE (2, "No gatekeeper found.");
      if (args.HasOption("require-gatekeeper"))
        return;
    }
  }

  PTRACE (2, "Waiting for incoming calls for \"" << endpoint.GetLocalUserName() << '"');;

  for (;;) 
    PThread::Current()->Sleep(5000);
}


///////////////////////////////////////////////////////////////

BOOL MyH323EndPoint::OnIncomingCall(H323Connection & _conn,
                                    const H323SignalPDU & setupPDU,
                                    H323SignalPDU &)
{
  return TRUE;
}

H323Connection * MyH323EndPoint::CreateConnection(unsigned callReference)
{
  return new MyH323Connection(*this, callReference, 0);
}

BOOL MyH323EndPoint::Initialise(PConfigArgs & args)
{
  if (args.HasOption('r')) {
    runCmd = args.GetOptionString('r');
    PTRACE (2,"Executing \"" << runCmd << "\" during message processing");
  }

 if (args.HasOption('H'))
    flags |= MyH323EndPoint::NoHangupAfterPlay;
  if (args.HasOption("loop"))
    flags |= MyH323EndPoint::LoopMessage;

 if (args.HasOption('m')) {
      if (PFile::Exists(args.GetOptionString('m'))) {
        ogmFile = args.GetOptionString('m');
      }
      else if (PFile::Exists(args.GetOptionString('m') + ".wav")) {
        ogmFile = args.GetOptionString('m') + ".wav";
      }
  }

  AddAllCapabilities(0, 0, "*");

  capabilities.Remove(args.GetOptionString('D').Lines());
  capabilities.Reorder(args.GetOptionString('P').Lines());

  PTRACE (2, "Codecs (in preference order):\n" << setprecision(2) << capabilities);

  return TRUE;
}


///////////////////////////////////////////////////////////////


MyH323Connection::MyH323Connection(MyH323EndPoint & _ep, unsigned callReference, unsigned options)
  : H323Connection(_ep, callReference, options), ep(_ep)
{
  ogmChannel = NULL;

  transmitCodecName = "none";

  PTRACE (2, "Opening connection");
}

MyH323Connection::~MyH323Connection()
{
  PTRACE (2,"Closing connection");

  PTime now;
  PTimeInterval interval = now - callStartTime;
  PString addr = GetControlChannel().GetRemoteAddress();
  addr.Delete (0,3); // strip ip$

  unsigned duration = (unsigned)((interval.GetMilliSeconds()+999)/1000);

  LogCall(duration, addr, calledParty, callerParty, callerID, transmitCodecName);

  if (ogmChannel != NULL)
    delete ogmChannel;
}

H323Connection::AnswerCallResponse
     MyH323Connection::OnAnswerCall(const PString & caller,
                                    const H323SignalPDU & setupPDU,
                                    H323SignalPDU & /*connectPDU*/)
{ 
  PTRACE (2, "Accepting call from " << caller);

  setupPDU.GetQ931().GetCalledPartyNumber (calledParty);
  setupPDU.GetQ931().GetCallingPartyNumber (callerParty);
  callerID = setupPDU.GetQ931().GetDisplayName ();

  return AnswerCallNow;
}

void MyH323Connection::PlayCommand ()
{
    if (ep.GetRunCmd().IsEmpty ()) 
	return;

    PString addr = GetControlChannel().GetRemoteAddress();
    addr.Delete (0,3); // strip ip$

    PStringArray argArray;
    PINDEX argCount = 0;
    argArray[argCount++] = addr;
    argArray[argCount++] = calledParty;
    argArray[argCount++] = callerParty;
    argArray[argCount++] = callerID;

    PTRACE (2, "Executing command \"" << ep.GetRunCmd () << "\" with args " << argArray);
     
    PPipeChannel cmd;
    int code = -1;
    
    if (!cmd.Open(ep.GetRunCmd (), argArray, PPipeChannel::ReadOnly)) {
       PTRACE(2, "Cannot execute command " << cmd);
       return;
    } 
    
    cmd.Execute();

    while (!cmd.eof()) {
      PString playFile;
      cmd >> playFile;
      
      if (playFile.IsEmpty ())
        continue;
      
      if (!PFile::Exists(playFile)) {
        PTRACE (1, "Error: cannot find file \"" << playFile << "\"");
      } 
      else {
        PTRACE (2, "Queue file \"" << playFile << "\"");
        ogmChannel->QueueFile(playFile);
      }
    }
      
    code = cmd.WaitForTermination();
    if (code >= 0) {
      PTRACE(2, "Command sucseeded");
    } else {
      PTRACE(2, "Command failed");
    }

    return;
}

BOOL MyH323Connection::OnCreateLogicalChannel(const H323Capability & capability,
                                            H323Channel::Directions dir,
                                            unsigned & errorCode)
{
    // Don't allow any channels than transmitter
    if (dir != H323Channel::IsTransmitter) 
      {
        errorCode = H245_OpenLogicalChannelReject_cause::e_dataTypeNotAvailable;
        return FALSE;
      }
    
    return TRUE;
}

BOOL MyH323Connection::OpenAudioChannel(BOOL isEncoding, 
                                        unsigned /* bufferSize */, 
                                        H323AudioCodec & codec)
{
  if (!isEncoding)
      return TRUE;
	
  PString codecName = codec.GetMediaFormat();

  static OpalMediaFormat::List list = H323PluginCodecManager::GetMediaFormats();
  if (list.GetValuesIndex(codecName) == P_MAX_INDEX) {
      PTRACE (1, "Unknown codec \"" << codecName);
      return FALSE;
  }
 
  PWaitAndSignal mutex(connMutex);

  if (ogmChannel == NULL) {
    ogmChannel = new PCM_OGMChannel(*this);
  }
  
  if (ep.GetHangupAfterPlay())
    ogmChannel->SetPlayOnce();

  if (ep.GetLoopMessage())
    ogmChannel->SetLoopMessage();

  transmitCodecName = codecName;
  codec.AttachChannel(ogmChannel, FALSE);

  if (PFile::Exists(ep.GetOgmFile())) {
    ogmChannel->QueueFile(ep.GetOgmFile ());      
  }
  PlayCommand (); 
  
  return TRUE;
}

///////////////////////////////////////////////////////////////

BOOL CheckWAVFileValid(PWAVFile *chan) {
  // Check the wave file header
  if (!chan->IsValid()) {
    PTRACE(1, chan->GetName() << " wav file header invalid");
    return FALSE;
  }

  // Check the wave file format
  if ((chan->GetFormat() != 0x01) ){
    PTRACE(1, chan->GetName() << " is not a PCM format wav file");
    PTRACE(1, "It is format " << chan->GetFormat() );
    return FALSE;
  }

  // Check the sample rate for PCM wave files
  if ( (chan->GetSampleRate() != 8000) ||
       (chan->GetChannels() != 1) ||
       (chan->GetSampleSize() != 16)
     ) {
    PTRACE(1, chan->GetName() << " is not a 16 Bit, Mono, 8000 Hz (8Khz) PCM wav file");
    PTRACE(1, "It is " << chan->GetSampleSize() << " bits, "
                       << (chan->GetChannels()==1 ? "mono " : "stereo ")
                       << chan->GetSampleRate() << " Hz");
    return FALSE;
  }

  return TRUE;
}

///////////////////////////////////////////////////////////////

PCM_OGMChannel::PCM_OGMChannel(MyH323Connection & _conn)
  : conn(_conn)
{
  silentCount = 50;         // wait 50 frames before playing the OGM
  closed        = FALSE;
  playOnce      = FALSE;
  loopMessage   = FALSE;
  finished      = FALSE;

  frameLen = frameOffs = 0;
}

BOOL PCM_OGMChannel::IsWAVFileValid(PWAVFile *chan) {
  // Check that this is a PCM wave file
  return CheckWAVFileValid(chan);
}

BOOL PCM_OGMChannel::Read(void * buffer, PINDEX amount)
{
  PWaitAndSignal mutex(chanMutex);

  // if the channel is closed, then return error
  if (closed)
    return FALSE;

  // Create the frame buffer using the amount of bytes the codec wants to
  // read. Different codecs use different read sizes.
  frameBuffer.SetMinSize(1024); //amount);

  // assume we are returning silence
  BOOL doSilence = TRUE;
  BOOL frameBoundary = FALSE;

  // if still outputting a frame from last time, then keep doing it
  if (frameOffs >= frameLen) {
    // if we are returning silence frames, then 
    if (silentCount > 0) 
      silentCount--;
    // if a channel is already open, don't do silence
    else if (GetBaseReadChannel() != NULL)
      doSilence = FALSE;
    // If not in silence and no existing channel, open a new file.
    else if (finished && playOnce) {
      PTRACE(1, "Clearing connection");
      conn.ClearCall();
    } else {
      PString * str = playQueue.Dequeue();
    
      if (str != NULL) {
	
        // check the file extension and open a .wav 
	
        if (((*str).Right(4)).ToLower() == ".wav") {

          PWAVFile *chan;
	  chan = new PWAVFile(*str, PFile::ReadOnly);
          
    	  if (!chan->IsOpen()) {
            PTRACE(1, "Cannot open file \"" << chan->GetName() << "\"");
            delete chan;
          } else if (!IsWAVFileValid(chan) ){
            PTRACE(1, chan->GetName() << " is not a valid wav file");
            delete chan;
            cerr << "wave file is invalid" << endl;
          } else {
            PTRACE(1, "Playing file \"" << chan->GetName() << "\"");
            totalData = 0;
            SetReadChannel(chan, TRUE);
            doSilence = FALSE;
            if (loopMessage) {
               PTRACE(1, "Looping file \"" << *str << "\"");
               playQueue.Enqueue(new PString(*str));
	    }
	  }
        } 
        delete str;  
      } 
    }
  
    // if not doing silence, try and read from the file
    if (!doSilence) {
  
      if (ReadFrame(amount)) {
        totalData += amount;  
      } else {
        PTRACE(1, "Finished playing " << totalData << " bytes");  
        PIndirectChannel::Close();
        // no silence
        doSilence = TRUE;
	if (playQueue.IsEmpty ()) {
	    finished = TRUE;
	    silentCount = 100;
	} else {
            silentCount = 5;   // always do 5 frames of silence after every file       
	}
      }
    } else {
      // start silence frame if required
      CreateSilenceFrame(amount);
    }
  }  

  frameBoundary = AdjustFrame(buffer, amount);
  // delay to synchronise to frame boundary
  if (frameBoundary)
    Synchronise(amount);

  return TRUE;
}

BOOL PCM_OGMChannel::Close()
{
  PWaitAndSignal mutex(chanMutex);
  closed = TRUE;
  PIndirectChannel::Close();
  return TRUE;
}

void PCM_OGMChannel::QueueFile(const PString & fn)
{
  PWaitAndSignal mutex(chanMutex);
  PTRACE(1, "Enqueueing file " << fn << " for playing");
  playQueue.Enqueue(new PString(fn));
}

BOOL PCM_OGMChannel::AdjustFrame(void * buffer, PINDEX amount)
{
  if ((frameOffs + amount) > frameLen) {
    cerr << "Reading past end of frame:offs=" << frameOffs << ",amt=" << amount << ",len=" << frameLen << endl;
    return TRUE;
  }

  memcpy(buffer, frameBuffer.GetPointer()+frameOffs, amount);
  frameOffs += amount;

  lastReadCount = amount;

  return frameOffs == frameLen;
}

void PCM_OGMChannel::Synchronise(PINDEX amount)
{
  ogm_delay.Delay(amount / 16);
}

void PCM_OGMChannel::CreateSilenceFrame(PINDEX amount)
{
  frameOffs = 0;
  frameLen  = amount;
  memset(frameBuffer.GetPointer(), 0, frameLen);
}

BOOL PCM_OGMChannel::ReadFrame(PINDEX amount)
{
  frameOffs = 0;
  frameLen  = amount;

  BOOL result = PIndirectChannel::Read(frameBuffer.GetPointer(), frameLen);

  // if we did not read a full frame of audio, fill the end of the
  // frame with zeros.
  PINDEX count = GetLastReadCount();
  if (count < frameLen)
    memset(frameBuffer.GetPointer()+count, 0, frameLen-count);

  return result;
}

///////////////////////////////////////////////////////////////
